#include "png.h"
#include "jpeglib.h"    
#include "jerror.h"
#include "elabmm.h"

int pam_write(char *fname, struct elmIMG *img)
{
	FILE *fp = NULL;
	int row = 0, col = 0, dep = 0;

	fp = fopen(fname, "wb");
	if (!fp)
		return -1;

	fprintf(fp, "P7\n");
	fprintf(fp, "WIDTH %d\n", img->width);
	fprintf(fp, "HEIGHT %d\n", img->height);
	fprintf(fp, "DEPTH %d\n", img->depth);
	fprintf(fp, "MAXVAL %d\n", img->maxval);
	fprintf(fp, "TUPLTYPE %s\n", img->depth > 1 ? (img->depth > 3 ? "RGB_ALPHA" : "RGB") : (img->maxval > 1 ? "GRAYSCALE" : "BLACKANDWHITE"));
	fprintf(fp, "ENDHDR\n");

	if (img->bytes == 1)
		for (row = 0; row < img->height; row++)
			fwrite(img->buffer + row * img->stride, 1, img->width * img->depth, fp);
	else if (img->bytes == 2)
		for (row = 0; row < img->height; row++) {
			for (col = 0; col < img->width; col++) {
				for (dep = 0; dep < img->depth; dep++) {
					unsigned short int temp = ((unsigned short int *) (img->buffer + row * img->stride))[col * img->depth + dep];
					temp = (temp >> 8) | (temp << 8);
					fwrite(&temp, 2, 1, fp);
				}
			}
		}

	fclose(fp);

	return 0;
}

int pam_read(char *fname, struct elmIMG *img)
{
	FILE *fp = NULL;
	int row = 0, col = 0, dep = 0;
	char temp[256] = "", temp1[256] = "";

	memset(img, 0, sizeof(struct elmIMG));

	fp = fopen(fname, "rb");
	if (!fp)
		return -1;

	fscanf(fp, "%s\n", temp);
	if (strncmp(temp, "P7", 2))
		return -1;

	do {
		fscanf(fp, "%255[^\n]\n", temp);
		if (!strncmp(temp, "WIDTH", 5))
			sscanf(temp, "WIDTH %d", &img->width);
		else if (!strncmp(temp, "HEIGHT", 6))
			sscanf(temp, "HEIGHT %d", &img->height);
		else if (!strncmp(temp, "DEPTH", 5))
			sscanf(temp, "DEPTH %d", &img->depth);
		else if (!strncmp(temp, "MAXVAL", 6))
			sscanf(temp, "MAXVAL %d", &img->maxval);
		else if (!strncmp(temp, "TUPLTYPE", 8)) {
			sscanf(temp, "TUPLTYPE %255s", temp1);
		}
	} while (strncmp(temp, "ENDHDR", 6));

	img->bytes = img->maxval > 255 ? 2 : 1;
	img->stride = 16 * ((img->width * img->depth * img->bytes + 15) / 16);

	img->buffer = (unsigned char *) calloc(img->height * img->stride, 1);
	if (img->buffer) {
		if (img->bytes == 1)
			for (row = 0; row < img->height; row++)
				fread(img->buffer + row * img->stride, 1, img->width * img->depth, fp);
		else if (img->bytes == 2)
			for (row = 0; row < img->height; row++) {
				for (col = 0; col < img->width; col++) {
					for (dep = 0; dep < img->depth; dep++) {
						unsigned short int temp;
						fread(&temp, 2, 1, fp);
						temp = (temp >> 8) | (temp << 8);
						((unsigned short int *) (img->buffer + row * img->stride))[col * img->depth + dep] = temp;
					}
				}
			}
	}

	fclose(fp);

	return 0;
}

int bmp_write(char *fname, struct elmIMG *img)
{
	int filesize = 0;
	/*                                 bfType   bfSize   bfRes1 bfRes2 bfOffBits */
	unsigned char bmpfileheader[14] = {'B','M', 0,0,0,0, 0,0,   0,0,   54,0,0,0};
	/*                                 biSize    biWidth  biHeight biPlanes biBitCount */
	unsigned char bmpinfoheader[40] = {40,0,0,0, 0,0,0,0, 0,0,0,0, 1,0,     0,0};
	unsigned char bmppad[3] = {0,0,0};
	unsigned char temp[4] = {0};

	FILE *fp = NULL;
	int row = 0, col = 0, dep = 0;

	if (img->depth != 1 && img->depth != 3 || img->bytes != 1)
		return -1;

	fp = fopen(fname, "wb");
	if (!fp)
		return -1;

	filesize = 54 + img->depth * img->width * img->height + (img->depth == 1 ? 256 * 4 : 0);

	bmpfileheader[2] = (unsigned char) filesize;
	bmpfileheader[3] = (unsigned char) (filesize>>8);
	bmpfileheader[4] = (unsigned char) (filesize>>16);
	bmpfileheader[5] = (unsigned char) (filesize>>24);

	bmpinfoheader[4] = (unsigned char) img->width;
	bmpinfoheader[5] = (unsigned char) (img->width>>8);
	bmpinfoheader[6] = (unsigned char) (img->width>>16);
	bmpinfoheader[7] = (unsigned char) (img->width>>24);
	bmpinfoheader[8] = (unsigned char) img->height;
	bmpinfoheader[9] = (unsigned char) (img->height>>8);
	bmpinfoheader[10] = (unsigned char) (img->height>>16);
	bmpinfoheader[11] = (unsigned char) (img->height>>24);

	bmpinfoheader[14] = img->depth == 1 ? 8 : (img->depth == 3 ? 24 : 32);
	bmpinfoheader[33] = img->depth == 1 ? 1 : 0;

	fwrite(bmpfileheader, 1, 14, fp);
	fwrite(bmpinfoheader, 1, 40, fp);

	for (row = 0; row < bmpinfoheader[33] * 256; row++) {
		temp[0] = temp[1] = temp[2] = row;
		fwrite(temp, 1, 4, fp);
	}

	for (row = img->height - 1; row >= 0; row--) {
		for (col = 0; col < img->width; col++) {
			for (dep = img->depth - 1; dep >= 0; dep--) {
				temp[0] = *(img->buffer + img->stride * row + img->depth * img->bytes * col + dep);
				fwrite(temp, 1, 1, fp);
			}
		}
		fwrite(bmppad, 1, (4 - (img->width * img->depth * img->bytes) % 4) % 4, fp);
	}

	fclose(fp);

	return 0;
}

int bmp_read(char *fname, struct elmIMG *img)
{
	FILE *fp = NULL;
	int row = 0, col = 0, dep = 0;
	/*                                 bfType bfSize   bfRes1 bfRes2 bfOffBits */
	unsigned char bmpfileheader[14] = {0,0,   0,0,0,0, 0,0,   0,0,   0,0,0,0};
	/*                                 biSize   biWidth  biHeight biPlanes biBitCount */
	unsigned char bmpinfoheader[40] = {0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,     0,0};
	unsigned short int biBitCount = 0;
	unsigned char bmppad[3] = {0,0,0};
	unsigned char temp[4] = {0};

	memset(img, 0, sizeof(struct elmIMG));

	fp = fopen(fname, "rb");
	if (!fp)
		return -1;

	fread(bmpfileheader, 1, 14, fp);
	if (bmpfileheader[0] != 'B' && bmpfileheader[1] != 'M')
		return -1;
	fread(bmpinfoheader, 1, 40, fp);

	img->width = (bmpinfoheader[7] << 24) | (bmpinfoheader[6] << 16) | (bmpinfoheader[5] << 8) | bmpinfoheader[4];
	img->height = (bmpinfoheader[11] << 24) | (bmpinfoheader[10] << 16) | (bmpinfoheader[9] << 8) | bmpinfoheader[8];
	biBitCount = (bmpinfoheader[15] << 8) | bmpinfoheader[14];

	if (biBitCount != 24)
		return -1;

	if (biBitCount == 24)
		img->depth = 3;
	else
		img->depth = 1;

	img->maxval = 255;
	img->bytes = img->maxval > 255 ? 2 : 1;
	img->stride = 16 * ((img->width * img->depth * img->bytes + 15) / 16);

	img->buffer = (unsigned char *) calloc(img->height * img->stride, 1);
	if (!img->buffer)
		return -1;

	for (row = img->height - 1; row >= 0; row--) {
		for (col = 0; col < img->width; col++) {
			for (dep = img->depth - 1; dep >= 0; dep--) {
				fread(temp, 1, 1, fp);
				*(img->buffer + img->stride * row + img->depth * img->bytes * col + dep) = temp[0];
			}
		}
		fread(bmppad, 1, (4 - (img->width * img->depth * img->bytes) % 4) % 4, fp);
	}

	return 0;
}

/* http://www.libpng.org/pub/png/book/chapter15.html */
int png_read(char *fname, struct elmIMG *img)
{
    unsigned char header[8];    // 8 is the maximum size that can be checked
	png_structp png_ptr;
	png_infop info_ptr;
	png_byte color_type;
	png_byte bit_depth;
	int number_of_passes;
	png_bytep *row_pointers;
	int y;

	/* open file and test for it being a png */
	FILE *fp = fopen(fname, "rb");
	if (!fp)
		return -1;
	fread(header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8))
		return -1;

    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
		return -1;

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
		return -1;

    /*if (setjmp(png_jmpbuf(png_ptr)))
		return -1;*/

    /* Expand any grayscale, RGB, or palette images to RGBA */
    png_set_expand(png_ptr);

    /* Reduce any 16-bits-per-sample images to 8-bits-per-sample */
    png_set_strip_16(png_ptr);

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

	memset(img, 0, sizeof(struct elmIMG));

    img->width = png_get_image_width(png_ptr, info_ptr);
    img->height = png_get_image_height(png_ptr, info_ptr);
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

	img->depth = color_type == PNG_COLOR_TYPE_GRAY ? 1 : (color_type == PNG_COLOR_TYPE_RGB ? 3 : 4);
	img->maxval = 255;
	img->bytes = img->maxval > 255 ? 2 : 1;
	img->stride = 16 * ((img->width * img->depth * img->bytes + 15) / 16);
	img->buffer = (unsigned char *) calloc(img->height * img->stride, 1);
	if (!img->buffer)
		return -1;

    /* read file */
    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * img->height);
    for (y = 0; y < img->height; y++)
		row_pointers[y] = img->buffer + y * img->stride;

    png_read_image(png_ptr, row_pointers);

	free(row_pointers);
    fclose(fp);

	if (png_ptr && info_ptr)
			png_destroy_write_struct(&png_ptr, &info_ptr);

	return 0;
}

/* http://www.libpng.org/pub/png/book/chapter15.html */
int png_write(char* file_name, struct elmIMG *img)
{
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep *row_pointers;
	int y;

	/* create file */
	FILE *fp = fopen(file_name, "wb");
	if (!fp)
		return -1;

	/* initialize stuff */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr)
		return -1;

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
		return -1;

	/*if (setjmp(png_jmpbuf(png_ptr)))
		return -1;*/

	png_init_io(png_ptr, fp);

	/* write header */
	/*if (setjmp(png_jmpbuf(png_ptr)))
		return -1;*/

	png_set_IHDR(png_ptr, info_ptr, img->width, img->height,
		8, img->depth == 3 ? PNG_COLOR_TYPE_RGB : (img->depth == 1 ? PNG_COLOR_TYPE_GRAY : PNG_COLOR_TYPE_RGB_ALPHA), PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	png_write_info(png_ptr, info_ptr);

	/* write bytes */
	/*if (setjmp(png_jmpbuf(png_ptr)))
		return -1;*/

    /* read file */
    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * img->height);
    for (y = 0; y < img->height; y++)
		row_pointers[y] = img->buffer + y * img->stride;
	png_write_image(png_ptr, row_pointers);

	/* end write */
	/*if (setjmp(png_jmpbuf(png_ptr)))
		return -1;*/

	png_write_end(png_ptr, NULL);

	/* cleanup heap allocation */
	free(row_pointers);
	fclose(fp);

	if (png_ptr && info_ptr)
        png_destroy_write_struct(&png_ptr, &info_ptr);

	return 0;
}

int jpg_read(char *fname, struct elmIMG *img)
{
	struct jpeg_decompress_struct info; //for our jpeg info
	struct jpeg_error_mgr err;          //the error handler
  
	FILE* fp = fopen(fname, "rb");  //open the file
	if (!fp)
		return -1;

	info.err = jpeg_std_error(&err);     
	jpeg_create_decompress(&info);   //fills info structure
  
	jpeg_stdio_src(&info, fp);    
	jpeg_read_header(&info, TRUE);   // read jpeg file header

	jpeg_start_decompress(&info);    // decompress the file

	//set width and height
	img->width = info.output_width;
	img->height = info.output_height;
	img->depth = info.num_components;
	if (img->depth != 3 && img->depth != 1)
		return -1;
	img->maxval = 255;
	img->bytes = img->maxval > 255 ? 2 : 1;
	img->stride = 16 * ((img->width * img->depth * img->bytes + 15) / 16);
	img->buffer = (unsigned char *) calloc(img->height * img->stride, 1);
	if (!img->buffer)
		return -1;

	//--------------------------------------------
	// read scanlines one at a time & put bytes 
	//    in jdata[] array. Assumes an RGB image
	//--------------------------------------------
	while (info.output_scanline < info.output_height) // loop
	{
		// Enable jpeg_read_scanlines() to fill our jdata array
		unsigned char *rowptr[1] = {img->buffer + info.output_scanline * img->stride};  
		jpeg_read_scanlines(&info, rowptr, 1);
	}

	jpeg_finish_decompress(&info);

	return 0;
}

int img_write(char *fname, struct elmIMG *img)
{
	if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".pam"))
		return pam_write(fname, img);
	else if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".bmp"))
		return bmp_write(fname, img);
	else if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".png"))
		return png_write(fname, img);
	else
		return -1;
}

int img_read(char *fname, struct elmIMG *img)
{
	if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".pam"))
		return pam_read(fname, img);
	else if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".bmp"))
		return bmp_read(fname, img);
	else if (strrchr(fname, '.') && !strcmp(strrchr(fname, '.'), ".png"))
		return png_read(fname, img);
	else if (strrchr(fname, '.') && (!strcmp(strrchr(fname, '.'), ".jpg") || !strcmp(strrchr(fname, '.'), ".jpeg")))
		return jpg_read(fname, img);
	else
		return -1;
}
